"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This package aims at providing an all-round usable RISC-V emulator and debugger

It contains everything needed to run assembly files, so you don't need any custom compilers or toolchains
"""

# to read package version:
import importlib.metadata

__author__ = "Anton Lydike <Anton@Lydike.com>"
__copyright__ = "Copyright 2023 Anton Lydike"
__version__ = importlib.metadata.version(__name__)
